/******************************************************************************************
 
    This program solves a wave equation using the explicit method.

    A wave equation is utt = c * c * uxx.
    The function f is the initial shape of wave.

    Time loop in this program consists of
    "Calculation part" and "Draw Part".

    Calculation part is controlled by the interval parameter INTV.
    This numerical result is visualized by using "g_bird_view_3D" of GLSC3D.
    Please note that this program set "g_def_scale_3D".
 
*******************************************************************************************/

#include<stdio.h>
#include<glsc3d_3.h>

#define N (100)
#define M (100)

double f(double x,double y)
{
    return 3 * exp(-((x - M_PI / 2)*(x - M_PI / 2) + (y  - 3 * M_PI / 2)*(y  - 3 * M_PI / 2))/ 0.1);
}

int main()
{
    double  x, y;
    
    double  Lx = 8 * M_PI;
    double  Ly = 8 * M_PI;
    double  Lz = 8 * M_PI;
    double  dx = Lx / N;
    double  dy = Ly / M;
    double  dt = 0.002;
    double  c = 1.0;
    
    double  u0[N + 2][M + 2];
    double  u1[N + 2][M + 2];
    double  u2[N + 2][M + 2];
    
    double  lambda_x = c * dt / dx;
    double  lambda_y = c * dt / dy;
    
    int     INTV = 100;
    
    g_init("Window", 600, 600);	//Pixel Size
    
    g_def_scale_3D_fix(0,                   //ID
                   -Lx*0.5, Lx*0.5,		//xmin,xmax
                   -Ly*0.5, Ly*0.5,     //ymin,ymax
                   -Lz*0.5, Lz*0.5,     //zmin,zmax
                   20.0, 20.0,          //Window (Left, Top) Position
                   560, 560);           //Window Size (x,y)
    
//    g_vision(0,                         //ID
//             0.0, -Ly, Lz,              //eye
//             0.0, 0.0, 1.0,             //up
//             1.25                       //zoom
//             );
    //Set initial calculation
    {
        //Step1
        for(int j = 1;j < M + 1; j ++)
        {
            y = (j - 0.5) * dy;
            for(int i = 1;i < N + 1; i ++)
            {
                x = (i - 0.5) * dx;
                u0[i][j] = f(x,y);
            }
        }
        //Step2
        for(int j = 1;j < M + 1; j ++)
        {
            u0[0][j] = -u0[1][j];
            u0[N + 1][j] = -u0[N][j];
        }
        for(int i = 0;i < N + 2; i ++)
        {
            u0[i][0] = -u0[i][1];
            u0[i][M + 1] = -u0[i][M];
        }
        //Step3
        for(int j = 1;j < M + 1; j ++)
        {
            y = (j - 0.5) * dy;
            for(int i = 1;i < N + 1; i ++)
            {
                x = (i - 0.5) * dx;
                u1[i][j] = u0[i][j] + lambda_x * lambda_x * (u0[i - 1][j] - 2 * u0[i][j] + u0[i + 1][j]) / 2
                + lambda_y * lambda_y * (u0[i][j - 1] - 2 * u0[i][j] + u0[i][j + 1]) / 2;
            }
        }
        //Step4
        for(int j = 1;j < M + 1; j ++)
        {
            u1[0][j] = -u1[1][j];
            u1[N + 1][j] = -u1[N][j];
        }
        for(int i = 0;i < N + 2; i ++)
        {
            u1[i][0] = -u1[i][1];
            u1[i][M + 1] = -u1[i][M];
        }
    }
    
    //////////// Start time loop ////////////
    for (int i_time = 0; ; i_time++) {

        //////////// Draw Part ////////////
        if (i_time%INTV == 0) {
            g_cls();					//Clear window
            g_sel_scale(0);				//Select Virtual scale
            g_line_color(0.0, 0.0, 0.0, 1.0);
            g_boundary();				//Draw Boundary
            
            g_area_color(1.0, 0.0, 0.0, 1.0);
            g_bird_view_3D(-Lx*0.5 + dx * 0.5, Lx*0.5 - dx * 0.5, -Ly*0.5 + dy * 0.5, Ly*0.5 - dy * 0.5, N+2, M+2, u0, G_NO, G_YES); //Bird View
            
            g_finish();					//flush Draw buffer
            g_sleep(0.01);				//Sleep 0.01 sec
        }
        
        //////////// Calculation Part ////////////
        {
            //Step5
            for(int j = 1;j < M + 1; j ++)
            {
                for(int i = 1;i < N + 1; i ++)
                {
                    u2[i][j] = 2 * u1[i][j] - u0[i][j]
                    + lambda_x * lambda_x *
                    (u1[i - 1][j] -2 * u1[i][j] + u1[i + 1][j])
                    + lambda_y * lambda_y *
                    (u1[i][j - 1] -2 * u1[i][j] + u1[i][j + 1]);
                }
            }
            //Step5
            for(int j = 1;j < M + 1; j ++)
            {
                for(int i = 1;i < N + 1; i ++)
                {
                    u0[i][j] = u1[i][j];
                }
            }
            //Step6
            for(int j = 1;j < M + 1; j ++)
            {
                u0[0][j] = -u0[1][j];
                u0[N + 1][j] = -u0[N][j];
            }
            for(int i = 0;i < N + 2; i ++)
            {
                u0[i][0] = -u0[i][1];
                u0[i][M + 1] = -u0[i][M];
            }
            //Step7
            for(int j = 1;j < M + 1; j ++)
            {
                for(int i = 1;i < N + 1; i ++)
                {
                    u1[i][j] = u2[i][j];
                }
            }
            //Step8
            for(int j = 1;j < M + 1; j ++)
            {
                u1[0][j] = -u1[1][j];
                u1[N + 1][j] = -u1[N][j];
            }
            for(int i = 0;i < N + 2; i ++)
            {
                u1[i][0] = -u1[i][1];
                u1[i][M + 1] = -u1[i][M];
            }
        }
    }
	return 0;
}
